@@ -10,11 +10,12 @@ use std::ops::Deref;
 use std::ptr::copy_nonoverlapping;
 use std::{cmp, result};
 
+use vhost_user_backend::bitmap::BitmapMmapRegion;
 use virtio_queue::DescriptorChain;
 use vm_memory::bitmap::{Bitmap, BitmapSlice};
 use vm_memory::{
-    Address, ByteValued, GuestMemory, GuestMemoryError, GuestMemoryMmap, GuestMemoryRegion, Le16,
-    Le32, Le64, VolatileMemory, VolatileMemoryError, VolatileSlice,
+    Address, ByteValued, GuestMemory, GuestMemoryError, GuestMemoryMmap, GuestMemoryRegion,
+    VolatileMemory, VolatileMemoryError, VolatileSlice,
 };
 
 use crate::file_traits::FileReadWriteAtVolatile;
@@ -184,7 +185,7 @@ impl<'a, B: BitmapSlice> DescriptorChainConsumer<'a, B> {
 /// Reader will skip iterating over descriptor chain when first writable
 /// descriptor is encountered.
 #[derive(Clone)]
-pub struct Reader<'a, B = ()> {
+pub struct Reader<'a, B = BitmapMmapRegion> {
     buffer: DescriptorChainConsumer<'a, B>,
 }
 
@@ -314,7 +315,7 @@ impl<'a, B: BitmapSlice> io::Read for Reader<'a, B> {
 /// Writer will start iterating the descriptors from the first writable one and will
 /// assume that all following descriptors are writable.
 #[derive(Clone)]
-pub struct Writer<'a, B = ()> {
+pub struct Writer<'a, B = BitmapMmapRegion> {
     buffer: DescriptorChainConsumer<'a, B>,
 }
 
@@ -435,39 +436,39 @@ pub enum DescriptorType {
     Writable,
 }
 
-#[derive(Copy, Clone, Debug, Default)]
-#[repr(C)]
-struct virtq_desc {
-    addr: Le64,
-    len: Le32,
-    flags: Le16,
-    next: Le16,
-}
-
-// Safe because it only has data and has no implicit padding.
-unsafe impl ByteValued for virtq_desc {}
-
-#[derive(Copy, Clone, Debug, Default)]
-#[repr(C)]
-struct virtq_avail {
-    flags: Le16,
-    idx: Le16,
-    ring: Le16,
-}
-
-// Safe because it only has data and has no implicit padding.
-unsafe impl ByteValued for virtq_avail {}
-
 #[cfg(test)]
 mod tests {
     use super::*;
     use virtio_queue::{Queue, QueueOwnedT, QueueT};
-    use vm_memory::{Bytes, GuestAddress};
+    use vm_memory::{Bytes, GuestAddress, Le16, Le32, Le64};
 
     const VIRTQ_DESC_F_NEXT: u16 = 0x1;
     const VIRTQ_DESC_F_WRITE: u16 = 0x2;
     const MAX_QUEUE_SIZE: u16 = 32768;
 
+    #[derive(Copy, Clone, Debug, Default)]
+    #[repr(C)]
+    struct virtq_desc {
+        addr: Le64,
+        len: Le32,
+        flags: Le16,
+        next: Le16,
+    }
+
+    // Safe because it only has data and has no implicit padding.
+    unsafe impl ByteValued for virtq_desc {}
+
+    #[derive(Copy, Clone, Debug, Default)]
+    #[repr(C)]
+    struct virtq_avail {
+        flags: Le16,
+        idx: Le16,
+        ring: Le16,
+    }
+
+    // Safe because it only has data and has no implicit padding.
+    unsafe impl ByteValued for virtq_avail {}
+
     /// Test utility function to create a descriptor chain in guest memory.
     pub fn create_descriptor_chain(
         memory: &GuestMemoryMmap,
