@@ -9,10 +9,13 @@ use passthrough::xattrmap::XattrMap;
 use std::collections::HashSet;
 use std::convert::{self, TryFrom, TryInto};
 use std::ffi::CString;
+use std::fs::File;
 use std::os::unix::io::{FromRawFd, RawFd};
 use std::path::Path;
 use std::str::FromStr;
-use std::sync::{Arc, RwLock};
+use std::sync::atomic::{AtomicBool, Ordering};
+use std::sync::{Arc, Mutex, RwLock};
+use std::thread::{self, JoinHandle};
 use std::time::Duration;
 use std::{env, error, fmt, io, process};
 use virtiofsd::idmap::{GidMap, UidMap};
@@ -22,6 +25,7 @@ use clap::{CommandFactory, Parser};
 use vhost::vhost_user::message::*;
 use vhost::vhost_user::Error::Disconnected;
 use vhost::vhost_user::{Backend, Listener};
+use vhost_user_backend::bitmap::BitmapMmapRegion;
 use vhost_user_backend::Error::HandleRequest;
 use vhost_user_backend::{VhostUserBackend, VhostUserDaemon, VringMutex, VringState, VringT};
 use virtio_bindings::bindings::virtio_config::*;
@@ -30,12 +34,14 @@ use virtio_bindings::bindings::virtio_ring::{
 };
 use virtio_queue::{DescriptorChain, QueueOwnedT};
 use virtiofsd::descriptor_utils::{Error as VufDescriptorError, Reader, Writer};
-use virtiofsd::filesystem::FileSystem;
-use virtiofsd::passthrough::{self, CachePolicy, InodeFileHandlesMode, PassthroughFs};
+use virtiofsd::filesystem::{FileSystem, SerializableFileSystem};
+use virtiofsd::passthrough::{
+    self, CachePolicy, InodeFileHandlesMode, MigrationMode, MigrationOnError, PassthroughFs,
+};
 use virtiofsd::sandbox::{Sandbox, SandboxMode};
 use virtiofsd::seccomp::{enable_seccomp, SeccompAction};
 use virtiofsd::server::Server;
-use virtiofsd::util::write_pid_file;
+use virtiofsd::util::{other_io_error, write_pid_file};
 use virtiofsd::{limits, oslib, Error as VhostUserFsError};
 use vm_memory::{
     ByteValued, GuestAddressSpace, GuestMemoryAtomic, GuestMemoryLoadGuard, GuestMemoryMmap, Le32,
@@ -43,7 +49,7 @@ use vm_memory::{
 use vmm_sys_util::epoll::EventSet;
 use vmm_sys_util::eventfd::EventFd;
 
-const QUEUE_SIZE: usize = 1024;
+const QUEUE_SIZE: usize = 32768;
 // The spec allows for multiple request queues. We currently only support one.
 const REQUEST_QUEUES: u32 = 1;
 // In addition to the request queue there is one high-prio queue.
@@ -51,6 +57,9 @@ const REQUEST_QUEUES: u32 = 1;
 // notification queue.
 const NUM_QUEUES: usize = REQUEST_QUEUES as usize + 1;
 
+type LoggedMemory = GuestMemoryMmap<BitmapMmapRegion>;
+type LoggedMemoryAtomic = GuestMemoryAtomic<LoggedMemory>;
+
 // The guest queued an available buffer for the high priority queue.
 const HIPRIO_QUEUE_EVENT: u16 = 0;
 // The guest queued an available buffer for the request queue.
@@ -61,6 +70,9 @@ const MAX_TAG_LEN: usize = 36;
 type Result<T> = std::result::Result<T, Error>;
 type VhostUserBackendResult<T> = std::result::Result<T, std::io::Error>;
 
+// The compiler warns that some wrapped values are never read, but they are in fact read by
+// `<Error as fmt::Display>::fmt()` via the derived `Debug`.
+#[allow(dead_code)]
 #[derive(Debug)]
 enum Error {
     /// Failed to create kill eventfd.
@@ -112,12 +124,12 @@ impl error::Error for Error {}
 
 impl convert::From<Error> for io::Error {
     fn from(e: Error) -> Self {
-        io::Error::new(io::ErrorKind::Other, e)
+        other_io_error(e)
     }
 }
 
 struct VhostUserFsThread<F: FileSystem + Send + Sync + 'static> {
-    mem: Option<GuestMemoryAtomic<GuestMemoryMmap>>,
+    mem: Option<LoggedMemoryAtomic>,
     kill_evt: EventFd,
     server: Arc<Server<F>>,
     // handle request from backend to frontend
@@ -139,7 +151,7 @@ impl<F: FileSystem + Send + Sync + 'static> Clone for VhostUserFsThread<F> {
     }
 }
 
-impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
+impl<F: FileSystem + SerializableFileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
     fn new(fs: F, thread_pool_size: usize) -> Result<Self> {
         let pool = if thread_pool_size > 0 {
             // Test that unshare(CLONE_FS) works, it will be called for each thread.
@@ -179,7 +191,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
     }
 
     fn return_descriptor(
-        vring_state: &mut VringState,
+        vring_state: &mut VringState<LoggedMemoryAtomic>,
         head_index: u16,
         event_idx: bool,
         len: usize,
@@ -210,7 +222,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
         }
     }
 
-    fn process_queue_pool(&self, vring: VringMutex) -> Result<bool> {
+    fn process_queue_pool(&self, vring: VringMutex<LoggedMemoryAtomic>) -> Result<bool> {
         let mut used_any = false;
         let atomic_mem = match &self.mem {
             Some(m) => m,
@@ -257,7 +269,10 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
         Ok(used_any)
     }
 
-    fn process_queue_serial(&self, vring_state: &mut VringState) -> Result<bool> {
+    fn process_queue_serial(
+        &self,
+        vring_state: &mut VringState<LoggedMemoryAtomic>,
+    ) -> Result<bool> {
         let mut used_any = false;
         let mem = match &self.mem {
             Some(m) => m.memory(),
@@ -265,7 +280,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
         };
         let mut vu_req = self.vu_req.clone();
 
-        let avail_chains: Vec<DescriptorChain<GuestMemoryLoadGuard<GuestMemoryMmap>>> = vring_state
+        let avail_chains: Vec<DescriptorChain<GuestMemoryLoadGuard<LoggedMemory>>> = vring_state
             .get_queue_mut()
             .iter(mem.clone())
             .map_err(|_| Error::IterateQueue)?
@@ -298,7 +313,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
     fn handle_event_pool(
         &self,
         device_event: u16,
-        vrings: &[VringMutex],
+        vrings: &[VringMutex<LoggedMemoryAtomic>],
     ) -> VhostUserBackendResult<()> {
         let idx = match device_event {
             HIPRIO_QUEUE_EVENT => {
@@ -319,7 +334,10 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
             // requests on the queue.
             loop {
                 vrings[idx].disable_notification().unwrap();
-                self.process_queue_pool(vrings[idx].clone())?;
+                // we can't recover from an error here, so let's hope it's transient
+                if let Err(e) = self.process_queue_pool(vrings[idx].clone()) {
+                    error!("processing the vring {idx}: {e}");
+                }
                 if !vrings[idx].enable_notification().unwrap() {
                     break;
                 }
@@ -335,7 +353,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
     fn handle_event_serial(
         &self,
         device_event: u16,
-        vrings: &[VringMutex],
+        vrings: &[VringMutex<LoggedMemoryAtomic>],
     ) -> VhostUserBackendResult<()> {
         let mut vring_state = match device_event {
             HIPRIO_QUEUE_EVENT => {
@@ -356,7 +374,10 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserFsThread<F> {
             // requests on the queue.
             loop {
                 vring_state.disable_notification().unwrap();
-                self.process_queue_serial(&mut vring_state)?;
+                // we can't recover from an error here, so let's hope it's transient
+                if let Err(e) = self.process_queue_serial(&mut vring_state) {
+                    error!("processing the vring: {e}");
+                }
                 if !vring_state.enable_notification().unwrap() {
                     break;
                 }
@@ -390,21 +411,35 @@ impl Default for VirtioFsConfig {
 
 unsafe impl ByteValued for VirtioFsConfig {}
 
-struct VhostUserFsBackend<F: FileSystem + Send + Sync + 'static> {
+struct PremigrationThread {
+    handle: JoinHandle<()>,
+    cancel: Arc<AtomicBool>,
+}
+
+struct VhostUserFsBackend<F: FileSystem + SerializableFileSystem + Send + Sync + 'static> {
     thread: RwLock<VhostUserFsThread<F>>,
+    premigration_thread: Mutex<Option<PremigrationThread>>,
+    migration_thread: Mutex<Option<JoinHandle<io::Result<()>>>>,
     tag: Option<String>,
 }
 
-impl<F: FileSystem + Send + Sync + 'static> VhostUserFsBackend<F> {
+impl<F: FileSystem + SerializableFileSystem + Send + Sync + 'static> VhostUserFsBackend<F> {
     fn new(fs: F, thread_pool_size: usize, tag: Option<String>) -> Result<Self> {
         let thread = RwLock::new(VhostUserFsThread::new(fs, thread_pool_size)?);
-        Ok(VhostUserFsBackend { thread, tag })
+        Ok(VhostUserFsBackend {
+            thread,
+            premigration_thread: None.into(),
+            migration_thread: None.into(),
+            tag,
+        })
     }
 }
 
-impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend for VhostUserFsBackend<F> {
-    type Bitmap = ();
-    type Vring = VringMutex;
+impl<F: FileSystem + SerializableFileSystem + Send + Sync + 'static> VhostUserBackend
+    for VhostUserFsBackend<F>
+{
+    type Bitmap = BitmapMmapRegion;
+    type Vring = VringMutex<LoggedMemoryAtomic>;
 
     fn num_queues(&self) -> usize {
         NUM_QUEUES
@@ -419,6 +454,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend for VhostUserFsBack
             | 1 << VIRTIO_RING_F_INDIRECT_DESC
             | 1 << VIRTIO_RING_F_EVENT_IDX
             | VhostUserVirtioFeatures::PROTOCOL_FEATURES.bits()
+            | VhostUserVirtioFeatures::LOG_ALL.bits()
     }
 
     fn protocol_features(&self) -> VhostUserProtocolFeatures {
@@ -426,7 +462,9 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend for VhostUserFsBack
             | VhostUserProtocolFeatures::BACKEND_REQ
             | VhostUserProtocolFeatures::BACKEND_SEND_FD
             | VhostUserProtocolFeatures::REPLY_ACK
-            | VhostUserProtocolFeatures::CONFIGURE_MEM_SLOTS;
+            | VhostUserProtocolFeatures::CONFIGURE_MEM_SLOTS
+            | VhostUserProtocolFeatures::LOG_SHMFD
+            | VhostUserProtocolFeatures::DEVICE_STATE;
 
         if self.tag.is_some() {
             protocol_features |= VhostUserProtocolFeatures::CONFIG;
@@ -468,11 +506,40 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend for VhostUserFsBack
         result
     }
 
+    fn acked_features(&self, features: u64) {
+        if features & VhostUserVirtioFeatures::LOG_ALL.bits() != 0 {
+            // F_LOG_ALL set: Prepare for migration (unless we're already doing that)
+            let mut premigration_thread = self.premigration_thread.lock().unwrap();
+            if premigration_thread.is_none() {
+                let cancel = Arc::new(AtomicBool::new(false));
+                let cloned_server = Arc::clone(&self.thread.read().unwrap().server);
+                let cloned_cancel = Arc::clone(&cancel);
+                let handle =
+                    thread::spawn(move || cloned_server.prepare_serialization(cloned_cancel));
+                *premigration_thread = Some(PremigrationThread { handle, cancel });
+            }
+        } else {
+            // F_LOG_ALL cleared: Migration cancelled, if any was ongoing
+            // (Note that this is our interpretation, and not said by the specification.  The back
+            // end might clear this flag also on the source side once the VM has been stopped, even
+            // before we receive SET_DEVICE_STATE_FD.  QEMU will clear F_LOG_ALL only when the VM
+            // is running, i.e. when the source resumes after a cancelled migration, which is
+            // exactly what we want, but it would be better if we had a more reliable way that is
+            // backed up by the spec.  We could delay cancelling until we receive a guest request
+            // while F_LOG_ALL is cleared, but that can take an indefinite amount of time.)
+            if let Some(premigration_thread) = self.premigration_thread.lock().unwrap().take() {
+                premigration_thread.cancel.store(true, Ordering::Relaxed);
+                // Ignore the result, we are cancelling anyway
+                let _ = premigration_thread.handle.join();
+            }
+        }
+    }
+
     fn set_event_idx(&self, enabled: bool) {
         self.thread.write().unwrap().event_idx = enabled;
     }
 
-    fn update_memory(&self, mem: GuestMemoryAtomic<GuestMemoryMmap>) -> VhostUserBackendResult<()> {
+    fn update_memory(&self, mem: LoggedMemoryAtomic) -> VhostUserBackendResult<()> {
         self.thread.write().unwrap().mem = Some(mem);
         Ok(())
     }
@@ -481,7 +548,7 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend for VhostUserFsBack
         &self,
         device_event: u16,
         evset: EventSet,
-        vrings: &[VringMutex],
+        vrings: &[VringMutex<LoggedMemoryAtomic>],
         _thread_id: usize,
     ) -> VhostUserBackendResult<()> {
         if evset != EventSet::IN {
@@ -504,6 +571,119 @@ impl<F: FileSystem + Send + Sync + 'static> VhostUserBackend for VhostUserFsBack
     fn set_backend_req_fd(&self, vu_req: Backend) {
         self.thread.write().unwrap().vu_req = Some(vu_req);
     }
+
+    fn set_device_state_fd(
+        &self,
+        direction: VhostTransferStateDirection,
+        phase: VhostTransferStatePhase,
+        file: File,
+    ) -> io::Result<Option<File>> {
+        // Our caller (vhost-user-backend crate) pretty much ignores error objects we return (only
+        // cares whether we succeed or not), so log errors here
+        if let Err(err) = self.do_set_device_state_fd(direction, phase, file) {
+            error!("Failed to initiate state (de-)serialization: {err}");
+            return Err(err);
+        }
+        Ok(None)
+    }
+
+    fn check_device_state(&self) -> io::Result<()> {
+        // Our caller (vhost-user-backend crate) pretty much ignores error objects we return (only
+        // cares whether we succeed or not), so log errors here
+        if let Err(err) = self.do_check_device_state() {
+            error!("Migration failed: {err}");
+            return Err(err);
+        }
+        Ok(())
+    }
+}
+
+impl<F: FileSystem + SerializableFileSystem + Send + Sync + 'static> VhostUserFsBackend<F> {
+    fn do_set_device_state_fd(
+        &self,
+        direction: VhostTransferStateDirection,
+        phase: VhostTransferStatePhase,
+        file: File,
+    ) -> io::Result<()> {
+        if phase != VhostTransferStatePhase::STOPPED {
+            return Err(io::Error::new(
+                io::ErrorKind::Unsupported,
+                format!("Transfer in phase {:?} is not supported", phase),
+            ));
+        }
+
+        let server = Arc::clone(&self.thread.read().unwrap().server);
+        let join_handle = match direction {
+            VhostTransferStateDirection::SAVE => {
+                // We should have a premigration thread that was started with `F_LOG_ALL`.  It
+                // should already be finished, but you never know.
+                let premigration_thread = self.premigration_thread.lock().unwrap().take();
+
+                thread::spawn(move || {
+                    if let Some(premigration_thread) = premigration_thread {
+                        // Let’s hope it’s finished.  Otherwise, we block migration downtime for a
+                        // bit longer, but there’s nothing we can do.
+                        premigration_thread.handle.join().map_err(|_| {
+                            other_io_error(
+                                "Failed to finalize serialization preparation".to_string(),
+                            )
+                        })?;
+                    } else {
+                        // If we don’t have a premigration thread, that either means migration was
+                        // cancelled at some point (i.e. F_LOG_ALL cleared; very unlikely and we
+                        // consider sending SET_DEVICE_STATE_FD afterwards a protocol violation),
+                        // or that there simply was no F_LOG_ALL at all.  QEMU doesn’t necessarily
+                        // do memory logging when snapshotting, and in such cases we have no choice
+                        // but to just run preserialization now.
+                        warn!(
+                            "Front-end did not announce migration to begin, so we failed to \
+                            prepare for it; collecting data now.  If you are doing a snapshot, \
+                            that is OK; otherwise, migration downtime may be prolonged."
+                        );
+                        server.prepare_serialization(Arc::new(AtomicBool::new(false)));
+                    }
+
+                    server.serialize(file).map_err(|e| {
+                        io::Error::new(e.kind(), format!("Failed to save state: {}", e))
+                    })
+                })
+            }
+
+            VhostTransferStateDirection::LOAD => {
+                if let Some(premigration_thread) = self.premigration_thread.lock().unwrap().take() {
+                    // Strange, but OK
+                    premigration_thread.cancel.store(true, Ordering::Relaxed);
+                    warn!("Cancelling serialization preparation because of incoming migration");
+                    let _ = premigration_thread.handle.join();
+                }
+
+                thread::spawn(move || {
+                    server.deserialize_and_apply(file).map_err(|e| {
+                        io::Error::new(e.kind(), format!("Failed to load state: {}", e))
+                    })
+                })
+            }
+        };
+
+        *self.migration_thread.lock().unwrap() = Some(join_handle);
+
+        Ok(())
+    }
+
+    fn do_check_device_state(&self) -> io::Result<()> {
+        let Some(migration_thread) = self.migration_thread.lock().unwrap().take() else {
+            // `check_device_state()` must follow a successful `set_device_state_fd()`, so this is
+            // a protocol violation
+            return Err(io::Error::new(
+                io::ErrorKind::InvalidInput,
+                "Front-end attempts to check migration state, but no migration has been done",
+            ));
+        };
+
+        migration_thread
+            .join()
+            .map_err(|_| other_io_error("Failed to join the migration thread"))?
+    }
 }
 
 fn parse_seccomp(src: &str) -> std::result::Result<SeccompAction, &'static str> {
@@ -574,7 +754,7 @@ fn parse_tag(tag: &str) -> Result<String> {
 )]
 struct Opt {
     /// Shared directory path
-    #[arg(long)]
+    #[arg(long, required_unless_present_any = &["compat_options", "print_capabilities"])]
     shared_dir: Option<String>,
 
     /// The tag that the virtio device advertises
@@ -643,9 +823,9 @@ struct Opt {
     /// - never: Never use file handles, always use O_PATH file descriptors.
     ///
     /// - prefer: Attempt to generate file handles, but fall back to O_PATH file descriptors where
-    /// the underlying filesystem does not support file handles.  Useful when there are various
-    /// different filesystems under the shared directory and some of them do not support file
-    /// handles.  ("fallback" is a deprecated alias for "prefer".)
+    ///   the underlying filesystem does not support file handles.  Useful when there are various
+    ///   different filesystems under the shared directory and some of them do not support file
+    ///   handles.  ("fallback" is a deprecated alias for "prefer".)
     ///
     /// - mandatory: Always use file handles, never fall back to O_PATH file descriptors.
     ///
@@ -660,6 +840,9 @@ struct Opt {
     #[arg(long, default_value = "auto")]
     cache: CachePolicy,
 
+    #[arg(long)]
+    allow_mmap: bool,
+
     /// Disable support for READDIRPLUS operations
     #[arg(long)]
     no_readdirplus: bool,
@@ -723,22 +906,97 @@ struct Opt {
     ///
     /// For example, :0:100000:65536: will map the 65536 host UIDs [100000, 165535]
     /// into the namespace as [0, 65535].
+    ///
+    /// Provide this argument multiple times to map multiple UID ranges.
     #[arg(long)]
-    uid_map: Option<UidMap>,
+    uid_map: Vec<UidMap>,
 
     /// Map a range of GIDs from the host into the namespace, given as
     /// :namespace_gid:host_gid:count:
     ///
     /// For example, :0:100000:65536: will map the 65536 host GIDs [100000, 165535]
     /// into the namespace as [0, 65535].
+    ///
+    /// Provide this argument multiple times to map multiple GID ranges.
     #[arg(long)]
-    gid_map: Option<GidMap>,
+    gid_map: Vec<GidMap>,
 
     /// Preserve O_NOATIME behavior, otherwise automatically clean up O_NOATIME flag to prevent
     /// potential permission errors when running in unprivileged mode (e.g., when accessing files
     /// without having ownership/capability to use O_NOATIME).
     #[arg(long = "preserve-noatime")]
     preserve_noatime: bool,
+
+    /// Defines how to perform migration, i.e. how to represent the internal state to the
+    /// destination, and how to obtain that representation.
+    ///
+    /// - find-paths: Obtain paths for all inodes indexed and opened by the guest, and transfer
+    ///   those paths to the destination.  To get those paths, we try to read the symbolic links in
+    ///   /proc/self/fd first; if that does not work, we will fall back to iterating through the
+    ///   shared directory (exhaustive search), enumerating all paths within.
+    ///
+    /// - file-handles: Pass file handles.  For this to work, source and destination instance must
+    ///   operate on exactly the same shared directory on the same filesystem (which may be a
+    ///   network filesystem, mounted on different hosts).  The destination instance must have the
+    ///   capability to open file handles, i.e. CAP_DAC_READ_SEARCH -- generally, this requires
+    ///   running virtiofsd as root and use `--modcaps=+dac_read_search`.
+    ///
+    /// This parameter is ignored on the destination side.
+    #[arg(long = "migration-mode", default_value = "find-paths")]
+    migration_mode: MigrationMode,
+
+    /// Controls how to respond to errors during migration.
+    ///
+    /// If any inode turns out not to be migrateable (either the source cannot serialize it, or the
+    /// destination cannot opened the serialized representation), the destination can react in
+    /// different ways:
+    ///
+    /// - abort: Whenever any error occurs, return a hard error to the vhost-user front-end (e.g.
+    ///          QEMU), aborting migration.
+    ///
+    /// - guest-error: Let migration finish, but the guest will be unable to access any of the
+    ///                affected inodes, receiving only errors.
+    ///
+    /// This parameter is ignored on the source side.
+    #[arg(long = "migration-on-error", default_value = "abort")]
+    migration_on_error: MigrationOnError,
+
+    /// Ensure that the migration destination opens the very same inodes as the source (only works
+    /// if source and destination use the same shared directory on the same filesystem).
+    ///
+    /// This option makes the source attach the respective file handle to each inode transferred
+    /// during migration.  Once the destination has (re-)opened the inode, it will generate the
+    /// file handle on its end, and compare, ensuring that it has opened the very same inode.
+    ///
+    /// (File handles are per-filesystem unique identifiers for inodes that, besides the inode ID,
+    /// also include a generation ID to protect against inode ID reuse.)
+    ///
+    /// Using this option protects against external parties renaming or replacing inodes
+    /// while migration is ongoing, which, without this option, can lead to data loss or
+    /// corruption, so it should always be used when other processes besides virtiofsd have write
+    /// access to the shared directory.  However, again, it only works if both source and
+    /// destination use the same shared directory.
+    ///
+    /// This parameter is ignored on the destination side.
+    #[arg(long = "migration-verify-handles")]
+    migration_verify_handles: bool,
+
+    /// Double-check the identity of inodes right before switching over to the destination,
+    /// potentially making migration more resilient when third parties have write access to the
+    /// shared directory.
+    ///
+    /// When representing migrated inodes using their paths relative to the shared directory,
+    /// double-check during switch-over to the destination that each path still matches the
+    /// respective inode, and on mismatch, try to correct it via the respective symlink in
+    /// /proc/self/fd.
+    ///
+    /// Because this option requires accessing each inode indexed or opened by the guest, it can
+    /// prolong the switch-over phase of migration (when both source and destination are paused)
+    /// for an indeterminate amount of time.
+    ///
+    /// This parameter is ignored on the destination side.
+    #[arg(long = "migration-confirm-paths")]
+    migration_confirm_paths: bool,
 }
 
 fn parse_compat(opt: Opt) -> Opt {
@@ -818,7 +1076,7 @@ fn parse_compat(opt: Opt) -> Opt {
 
     if let Some(compat_options) = opt.compat_options.as_ref() {
         for line in compat_options {
-            for option in line.to_string().split(',') {
+            for option in line.split(',') {
                 if option.contains('=') {
                     parse_tuple(&mut clean_opt, option);
                 } else {
@@ -833,7 +1091,11 @@ fn parse_compat(opt: Opt) -> Opt {
 
 fn print_capabilities() {
     println!("{{");
-    println!("  \"type\": \"fs\"");
+    println!("  \"type\": \"fs\",");
+    println!("  \"features\": [");
+    println!("    \"migrate-precopy\",");
+    println!("    \"separate-options\"");
+    println!("  ]");
     println!("}}");
 }
 
@@ -961,20 +1223,6 @@ fn drop_capabilities(inode_file_handles: InodeFileHandlesMode, modcaps: Option<S
     }
 }
 
-fn has_noatime_capability() -> bool {
-    // We may not have all permissions/capabilities to use O_NOATIME with all the exported files if
-    // we are running as unprivileged user and without any sandbox (e.g., --sandbox=none).
-    //
-    // Provide this helper function to check this particular case.
-    let uid = unsafe { libc::geteuid() };
-    let cap = capng::name_to_capability("FOWNER").unwrap_or_else(|err| {
-        error!("could not get capability FOWNER: {}", err);
-        process::exit(1);
-    });
-
-    uid == 0 || capng::have_capability(capng::Type::EFFECTIVE, cap)
-}
-
 fn main() {
     let opt = parse_compat(Opt::parse());
 
@@ -1000,6 +1248,13 @@ fn main() {
             process::exit(1);
         }
     };
+
+    let shadir_path = Path::new(shared_dir);
+    if !shadir_path.is_dir() && !shadir_path.is_file() {
+        error!("{shared_dir} does not exist");
+        process::exit(1);
+    }
+
     if opt.compat_foreground {
         warn!("Use of deprecated flag '-f': This flag has no effect, please remove it");
     }
@@ -1052,6 +1307,19 @@ fn main() {
                 opt.socket.as_ref().unwrap() // safe to unwrap because clap ensures either --socket or --socket-path are passed
             });
 
+            let socket_parent_dir = Path::new(socket).parent().unwrap_or_else(|| {
+                error!("Invalid socket file name");
+                process::exit(1);
+            });
+
+            if !socket_parent_dir.as_os_str().is_empty() && !socket_parent_dir.exists() {
+                error!(
+                    "{} does not exist or is not a directory",
+                    socket_parent_dir.to_string_lossy()
+                );
+                process::exit(1);
+            }
+
             let pid_file_name = socket.to_owned() + ".pid";
             let pid_file_path = Path::new(pid_file_name.as_str());
             let pid_file = write_pid_file(pid_file_path).unwrap_or_else(|error| {
@@ -1106,7 +1374,7 @@ fn main() {
 
     // Enter the sandbox, from this point the process will be isolated (or not)
     // as chosen in '--sandbox'.
-    sandbox.enter().unwrap_or_else(|error| {
+    let listener = sandbox.enter(listener).unwrap_or_else(|error| {
         error!("Error entering sandbox: {}", error);
         process::exit(1)
     });
@@ -1129,7 +1397,12 @@ fn main() {
         killpriv_v2,
         security_label: opt.security_label,
         posix_acl: opt.posix_acl,
-        clean_noatime: !opt.preserve_noatime && !has_noatime_capability(),
+        clean_noatime: !opt.preserve_noatime,
+        allow_mmap: opt.allow_mmap,
+        migration_on_error: opt.migration_on_error,
+        migration_verify_handles: opt.migration_verify_handles,
+        migration_confirm_paths: opt.migration_confirm_paths,
+        migration_mode: opt.migration_mode,
         ..Default::default()
     };
 
